import json
import logging
from typing import (
    Optional,
    Type,
    Tuple,
)

from langchain_core.callbacks import CallbackManagerForToolRun
from pydantic import Field, model_validator, BaseModel
from typing_extensions import Self

from ttyg.graphdb import GraphDBAutocompleteStatus, GraphDBRdfRankStatus
from ttyg.utils import timeit
from .base import BaseGraphDBTool


class AutocompleteSearchTool(BaseGraphDBTool):
    """
    Tool, which uses GraphDB Autocomplete index to search for IRIs by name and class.
    The agent generates the autocomplete search query and the target class, which are expanded in the SPARQL template.
    """

    class SearchInput(BaseModel):
        query: str = Field(description="autocomplete search query")
        result_class: Optional[str] = Field(
            description="Optionally, filter the results by class. ",
            default=None,
        )
        limit: Optional[int] = Field(description="limit the results", default=10, ge=1)

    name: str = "autocomplete_search"
    description: str = "Discover IRIs by searching their names and getting results in order of relevance."
    args_schema: Type[BaseModel] = SearchInput
    response_format: str = "content_and_artifact"
    sparql_query_template: str = """PREFIX rank: <http://www.ontotext.com/owlim/RDFRank#>
    PREFIX auto: <http://www.ontotext.com/plugins/autocomplete#>
    SELECT ?iri ?name ?rank {{
        ?iri auto:query "{query}" ;
            {property_path} ?name ;{filter_clause}
            rank:hasRDFRank5 ?rank.
    }}
    ORDER BY DESC(?rank)
    LIMIT {limit}"""
    property_path: str = Field(
        default="<http://www.w3.org/2000/01/rdf-schema#label>",
        examples=[
            "<http://www.w3.org/2000/01/rdf-schema#label> | <http://schema.org/name>",
        ],
    )

    @model_validator(mode="after")
    def graphdb_config(self) -> Self:
        autocomplete_status = self.graph.get_autocomplete_status()
        if autocomplete_status != GraphDBAutocompleteStatus.READY:
            logging.warning(
                f"The Autocomplete index status of the repository is \"{autocomplete_status.name}\". "
                f"It's recommended the status to be READY in order to use the Autocomplete search tool."
            )

        rdf_rank_status = self.graph.get_rdf_rank_status()
        if rdf_rank_status != GraphDBRdfRankStatus.COMPUTED:
            logging.warning(
                f"The RDF Rank status of the repository is \"{rdf_rank_status.name}\". "
                f"It's recommended the status to be COMPUTED in order to use the Autocomplete search tool."
            )
        return self

    @timeit
    def _run(
            self,
            query: str,
            limit: Optional[int] = 10,
            result_class: Optional[str] = None,
            run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Tuple[str, str]:
        query = self.sparql_query_template.format(
            query=query,
            property_path=self.property_path,
            filter_clause=f" a {result_class} ;" if result_class else "",
            limit=limit,
        )
        logging.debug(f"Searching with autocomplete query {query}")
        query_results, query = self.graph.eval_sparql_query(query)
        return json.dumps(query_results, indent=2), query



import time
from typing import Any

from langchain_core.messages import AIMessage, ToolMessage, ToolCall
from langchain_core.messages.ai import UsageMetadata
from langgraph.graph.state import CompiledStateGraph


def run_agent_for_evaluation(
        agent: CompiledStateGraph,
        question_id: str,
        input_: dict[str, Any] | Any,
) -> dict[str, Any]:
    """
    Runs an agent over a given question and returns the output as expected by the evaluation library https://github.com/Ontotext-AD/graphrag-eval.
    :param agent: the agent
    :type agent: CompiledStateGraph
    :param question_id: the unique question id for tracing purposes
    :type question_id: str
    :param input_: the input (question) passed to the agent
    :type input_: dict[str, Any] | Any
    :return: check https://github.com/Ontotext-AD/graphrag-eval documentation for the expected keys in the result
    :rtype: dict[str, Any]
    """
    try:
        sum_input_tokens, sum_output_tokens, sum_total_tokens = 0, 0, 0
        tools_calls, tools_outputs = [], dict()

        start = time.time()
        output = agent.invoke(input_)
        elapsed_sec = time.time() - start

        for message in output["messages"]:
            if isinstance(message, AIMessage):
                tool_calls: list[ToolCall] = message.tool_calls
                for tool_call in tool_calls:
                    tools_calls.append({
                        "name": tool_call["name"],
                        "args": tool_call["args"],
                        "id": tool_call["id"],
                    })
                usage_metadata: UsageMetadata = message.usage_metadata
                sum_input_tokens += usage_metadata["input_tokens"]
                sum_output_tokens += usage_metadata["output_tokens"]
                sum_total_tokens += usage_metadata["total_tokens"]
            elif isinstance(message, ToolMessage):
                tools_outputs[message.tool_call_id] = {
                    "status": message.status,
                    "output": message.content
                }

        for tool_call in tools_calls:
            tool_call.update(tools_outputs[tool_call["id"]])

        return {
            "question_id": question_id,
            "input_tokens": sum_input_tokens,
            "output_tokens": sum_output_tokens,
            "total_tokens": sum_total_tokens,
            "elapsed_sec": elapsed_sec,
            "actual_steps": tools_calls,
            "actual_answer": output["messages"][-1].content
        }
    except Exception as e:
        return {
            "question_id": question_id,
            "error": str(e),
            "status": "error",
        }



import logging
import time
from functools import wraps


def timeit(func):
    """
    Decorator function, which measures the elapsed time for calling the wrapped function and logs it.
    :param func: the wrapped function
    :return: the value returned from the wrapped function
    """

    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = (end_time - start_time) * 1000
        logging.debug(f"Finished {func.__qualname__} in {total_time:.2f} millis")
        return result

    return timeit_wrapper



import json
import logging
from typing import (
    Optional,
    ClassVar,
    Type,
    Tuple,
)

from langchain_core.callbacks import CallbackManagerForToolRun
from pydantic import Field, model_validator, BaseModel
from typing_extensions import Self

from ttyg.utils import timeit
from .base import BaseGraphDBTool


class RetrievalQueryTool(BaseGraphDBTool):
    """
    Tool, which uses GraphDB ChatGPT Retrieval Plugin Connector.
    ChatGPT Retrieval Plugin Connector must exist in order to use this tool.
    The agent generates the search query, which is expanded in the SPARQL template.
    """

    class SearchInput(BaseModel):
        query: str = Field(description="text query")
        limit: Optional[int] = Field(description="limit the results", default=5, ge=1)
        score: Optional[float] = Field(description="filter the results by score", default=0, ge=0, le=1)

    min_graphdb_version: ClassVar[str] = "10.4"
    name: str = "retrieval_search"
    description: str = "Query the vector database to retrieve relevant pieces of documents."
    args_schema: Type[BaseModel] = SearchInput
    response_format: str = "content_and_artifact"
    sparql_query_template: str = """PREFIX retr: <http://www.ontotext.com/connectors/retrieval#>
    PREFIX retr-inst: <http://www.ontotext.com/connectors/retrieval/instance#>
    SELECT * {{
        [] a retr-inst:{connector_name} ;
            retr:query "{query}" ;
            retr:limit {limit} ;
            retr:entities ?entity .
        ?entity retr:snippets _:s ;
            retr:score ?score;
        _:s retr:snippetField ?field ;
            retr:snippetText ?text .
        FILTER (?score > {score})
    }}"""
    connector_name: str

    @model_validator(mode="after")
    def check_retrieval_connector_exists(self) -> Self:
        if not self.graph.retrieval_connector_exists(self.connector_name):
            logging.warning(
                f"ChatGPT Retrieval connector with name \"{self.connector_name}\" doesn't exist."
            )
        return self

    @timeit
    def _run(
            self,
            query: str,
            limit: Optional[int] = 5,
            score: Optional[float] = 0,
            run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Tuple[str, str]:
        query = self.sparql_query_template.format(
            connector_name=self.connector_name,
            query=query,
            limit=limit,
            score=score,
        )
        logging.debug(f"Searching with retrieval query {query}")
        query_results = self.graph.eval_sparql_query(query, validation=False)
        return json.dumps(query_results, indent=2), query



import logging
import time
from typing import Any

from langchain_core.runnables import RunnableConfig
from langgraph.graph.state import CompiledStateGraph


def run_agent(
        agent: CompiledStateGraph,
        input_: dict[str, Any] | Any,
        config: RunnableConfig | None = None,
        last_message_id: str = None,
) -> str:
    sum_input_tokens, sum_output_tokens, sum_total_tokens = 0, 0, 0

    start = time.time()
    for s in agent.stream(input_, config, stream_mode="values"):
        messages = s["messages"]
        for message in reversed(messages):
            if message.id == last_message_id:
                break

            message.pretty_print()
            if hasattr(message, "usage_metadata"):
                usage_metadata = message.usage_metadata
                input_tokens, output_tokens, total_tokens = usage_metadata["input_tokens"], usage_metadata[
                    "output_tokens"], usage_metadata["total_tokens"]
                sum_input_tokens += input_tokens
                sum_output_tokens += output_tokens
                sum_total_tokens += total_tokens
                logging.debug(
                    f"Usage: input tokens: {input_tokens}, "
                    f"output tokens: {output_tokens}, "
                    f"total tokens: {total_tokens}")

        last_message_id = messages[-1].id

    logging.debug(
        f"Total usage: input tokens: {sum_input_tokens}, "
        f"output tokens: {sum_output_tokens}, "
        f"total tokens: {sum_total_tokens}"
    )
    logging.debug(
        f"Elapsed time: {time.time() - start:.2f} seconds"
    )
    return last_message_id



from datetime import datetime, timezone
from typing import Optional

from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.tools import BaseTool

from ttyg.utils import timeit


class NowTool(BaseTool):
    """
    Tool, which returns the current UTC date time in yyyy-mm-ddTHH:MM:SS format
    """

    name: str = "now"
    description: str = "Returns the current UTC date time in yyyy-mm-ddTHH:MM:SS format. Do not reuse responses."

    @timeit
    def _run(
            self,
            run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        return datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S")



import json
import logging
from typing import (
    Optional,
    ClassVar,
    Type,
    Tuple,
)

from langchain_core.callbacks import CallbackManagerForToolRun
from pydantic import Field, model_validator, BaseModel
from typing_extensions import Self

from ttyg.utils import timeit
from .base import BaseGraphDBTool


class SimilaritySearchQueryTool(BaseGraphDBTool):
    """
    Tool, which uses GraphDB Similarity Index.
    Similarity Index must exist in order to use this tool.
    The agent generates the similarity search term, which is expanded in the SPARQL template.
    """

    class SearchInput(BaseModel):
        query: str = Field(description="FTS search query")

    min_graphdb_version: ClassVar[str] = "8.7"
    name: str = "similarity_search"
    description: str = "Query GraphDB by full-text search and return a subgraph of RDF triples."
    args_schema: Type[BaseModel] = SearchInput
    response_format: str = "content_and_artifact"
    sparql_query_template: str = """PREFIX sim: <http://www.ontotext.com/graphdb/similarity/>
    PREFIX sim-index: <http://www.ontotext.com/graphdb/similarity/instance/>
    DESCRIBE ?documentID {{
        SELECT DISTINCT ?documentID {{
            ?search a sim-index:{index_name} ;
                sim:searchTerm "{query}";
                sim:documentResult ?result .
            ?result sim:value ?documentID ;
                sim:score ?score.
            FILTER(?score >= {similarity_score_threshold})
        }}
        ORDER BY DESC(?score)
        LIMIT {limit}
    }}"""
    limit: int = Field(default=10, ge=1)
    similarity_score_threshold: float = Field(default=0.6, ge=0)
    index_name: str

    @model_validator(mode="after")
    def check_similarity_index_exists(self) -> Self:
        if not self.graph.similarity_index_exists(self.index_name):
            logging.warning(
                f"Similarity index with name \"{self.index_name}\" doesn't exist."
            )
        return self

    @timeit
    def _run(
            self,
            query: str,
            run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Tuple[str, str]:
        query = self.sparql_query_template.format(
            index_name=self.index_name,
            query=query,
            limit=self.limit,
            similarity_score_threshold=self.similarity_score_threshold,
        )
        logging.debug(f"Searching with similarity query {query}")
        query_results = self.graph.eval_sparql_query(query, validation=False)
        return json.dumps(query_results, indent=2), query



import logging
import re
from enum import Enum
from functools import cached_property
from typing import Optional, Tuple, Any

import pyparsing
import requests
from SPARQLWrapper import SPARQLWrapper, Wrapper, JSON, TURTLE
from rdflib.plugins import sparql
from requests import Response


class GraphDBRdfRankStatus(Enum):
    CANCELED = "CANCELED"
    COMPUTED = "COMPUTED"
    COMPUTING = "COMPUTING"
    EMPTY = "EMPTY"
    ERROR = "ERROR"
    OUTDATED = "OUTDATED"
    CONFIG_CHANGED = "CONFIG_CHANGED"


class GraphDBAutocompleteStatus(Enum):
    READY = "READY"
    READY_CONFIG = "READY_CONFIG"
    ERROR = "ERROR"
    NONE = "NONE"
    BUILDING = "BUILDING"
    CANCELED = "CANCELED"


class GraphDB:
    """Ontotext GraphDB https://graphdb.ontotext.com/ Client"""

    def __init__(
            self,
            base_url: str,
            repository_id: str,
            connect_timeout: int = 2,
            read_timeout: int = 10,
            sparql_timeout: int = 15,
            auth_header: Optional[str] = None,
    ):
        """
        Initializes a GraphDB Client.

        :param base_url : GraphDB Base URL
        :type base_url: str
        :param repository_id: GraphDB Repository ID
        :type repository_id: str
        :param connect_timeout: connect timeout in seconds for calls to GraphDB REST API, default = 2
        :type connect_timeout: int
        :param read_timeout: read timeout in seconds for calls to GraphDB REST API, default = 10
        :type read_timeout: int
        :param sparql_timeout: timeout in seconds for calls to the SPARQL endpoint, default = 15
        :type sparql_timeout: int
        :param auth_header: optional, the value of the "Authorization" header to pass to GraphDB, if it's secured
        :type auth_header: Optional[str]
        """
        self.__base_url = base_url
        self.__repository_id = repository_id
        self.__connect_timeout = connect_timeout
        self.__read_timeout = read_timeout
        self.__sparql_wrapper = SPARQLWrapper(f"{base_url}/repositories/{repository_id}")
        self.__sparql_wrapper.setTimeout(sparql_timeout)
        self.__auth_header = auth_header
        if self.__auth_header:
            self.__sparql_wrapper.addCustomHttpHeader(
                "Authorization", self.__auth_header
            )

        self.__check_connectivity()

    def __check_connectivity(self):
        self.eval_sparql_query(query="ASK {?s ?p ?o}", validation=False)

    def __get_request(self, url: str, headers: dict, params=None) -> Response:
        if self.__auth_header:
            headers["Authorization"] = self.__auth_header

        response = requests.get(
            url,
            params=params,
            headers=headers,
            timeout=(self.__connect_timeout, self.__read_timeout),
        )
        response.raise_for_status()
        return response

    def get_known_prefixes(self) -> dict[str, str]:
        """
        Fetch all namespace declaration info available in the repository.

        :return: dictionary of prefix - namespace pairs,
        which are returned by GraphDB /repositories/{repositoryID}/namespaces endpoint
        :rtype: dict[str, str]
        """
        response = self.__get_request(
            f"{self.__base_url}/repositories/{self.__repository_id}/namespaces",
            headers={
                "Accept": "application/sparql-results+json",
            },
        )
        namespaces = dict()
        for binding in response.json()["results"]["bindings"]:
            namespaces[binding["prefix"]["value"]] = binding["namespace"]["value"]
        return namespaces

    def fts_is_enabled(self) -> bool:
        """
        Checks if the full-text search (FTS) is enabled
        using the GraphDB REST API /rest/repositories/{repository_id} endpoint.

        :return: True, if full-text search (FTS) is enabled; False, otherwise
        :rtype: bool
        """
        response = self.__get_request(
            f"{self.__base_url}/rest/repositories/{self.__repository_id}",
            headers={
                "Accept": "application/json",
            },
        )
        return response.json()["params"]["enableFtsIndex"]["value"].lower() == "true"

    def get_autocomplete_status(self) -> GraphDBAutocompleteStatus:
        """
        Returns the status of the autocomplete index for the repository.

        :rtype: GraphDBAutocompleteStatus
        """
        sparql_result, _ = self.eval_sparql_query(
            "PREFIX auto: <http://www.ontotext.com/plugins/autocomplete#> SELECT ?status { ?s auto:status ?status }",
            validation=False
        )
        return GraphDBAutocompleteStatus[sparql_result["results"]["bindings"][0]["status"]["value"]]

    def similarity_index_exists(self, index_name: str) -> bool:
        """
        Checks if a similarity index with the provided name exists
        using the GraphDB REST API /rest/similarity endpoint.

        :param index_name: the similarity index name
        :type index_name: str
        :return: True, if the index exists; False, otherwise
        :rtype: bool
        """
        response = self.__get_request(
            f"{self.__base_url}/rest/similarity",
            headers={
                "Accept": "application/json",
                "X-GraphDB-Repository": self.__repository_id,
            },
        )
        return index_name in {index["name"] for index in response.json()}

    def retrieval_connector_exists(self, connector_name: str) -> bool:
        """
        Checks if a ChatGPT Retrieval Plugin Connector with the provided name exists
        using the GraphDB REST API /rest/connectors/existing endpoint.

        :param connector_name: the connector name
        :type connector_name: str
        :return: True, if the connector exists; False, otherwise
        :rtype: bool
        """
        response = self.__get_request(
            f"{self.__base_url}/rest/connectors/existing",
            params={"prefix": "http://www.ontotext.com/connectors/retrieval#"},
            headers={
                "Accept": "application/json",
                "X-GraphDB-Repository": self.__repository_id,
            },
        )
        return connector_name in {connector["name"] for connector in response.json()}

    @cached_property
    def version(self) -> str:
        """
        :return: the GraphDB server version
        :rtype: str
        """
        response = self.__get_request(
            f"{self.__base_url}/rest/info/version",
            headers={
                "Accept": "application/json",
            },
        )
        return response.json()["productVersion"]

    def get_rdf_rank_status(self) -> GraphDBRdfRankStatus:
        """
        Returns the status of GraphDB RDF rank for the repository.

        :rtype: GraphDBRdfRankStatus
        """
        sparql_result, _ = self.eval_sparql_query(
            "PREFIX rank: <http://www.ontotext.com/owlim/RDFRank#> SELECT ?status { ?s rank:status ?status }",
            validation=False
        )
        return GraphDBRdfRankStatus[sparql_result["results"]["bindings"][0]["status"]["value"]]

    def __validate_query(self, query: str) -> str:
        """
        Validates a given SPARQL query and corrects its prefixes, if possible.

        :param query: SPARQL query
        :type query: str
        :return updated SPARQL query, where:
         - missing prefixes are added automatically, if they appear in the declaration info available in the repository
         - prefix definitions, which values differ from the declaration info available in the repository,
         are automatically corrected
        :rtype: str
        :raises ValueError, if:
        - the SPARQL query syntax is wrong
        - the query is an update SPARQL query
        - the query uses prefixes, which are not defined in the query and don't appear in
        the declaration info available in the repository
        - one or more IRIs used in the query are not stored in repository
        """
        parsed_query = self.__parse_query(query)
        prefix_part = str(parsed_query[0])
        query_part = str(parsed_query[1:])

        defined_prefixes = self.__get_defined_prefixes(prefix_part)
        known_prefixes = self.get_known_prefixes()

        query = self.__correct_wrong_prefixes(defined_prefixes, known_prefixes, query)

        prefixed_iris = self.__get_prefixed_iris(query_part)

        query = self.__add_missing_prefixes(defined_prefixes, known_prefixes, prefixed_iris, query)

        self.__validate_iris_are_stored(defined_prefixes, prefixed_iris, query_part)

        return query

    @staticmethod
    def __parse_query(query: str) -> pyparsing.results.ParseResults:
        """
`       Parses a given SPARQL query.
        If the query is an update SPARQL query, an exception is thrown, as we expect only read queries.

        :param query: SPARQL query
        :type query: str
        :return: the parsed SPARQL query
        :rtype: pyparsing.results.ParseResults
        :raises ValueError, if the SPARQL query syntax is wrong or the query is an update SPARQL query
        """
        try:
            return sparql.parser.parseQuery(query)
        except pyparsing.exceptions.ParseException as e:
            raise ValueError(e)

    @staticmethod
    def __get_defined_prefixes(prefix_part: str) -> dict[str, str]:
        """
        Returns the defined prefixes in the SPARQL query
        :param prefix_part: the prefix part of the parsed query
        :type prefix_part: str
        :return: the defined prefix - namespace pairs as dictionary
        :rtype: dict[str, str]
        """
        return dict(
            re.findall(
                r"PrefixDecl_\{'prefix': '(.+?)', 'iri': rdflib\.term\.URIRef\('(.+?)'\)}",
                prefix_part
            )
        )

    @staticmethod
    def __correct_wrong_prefixes(
            defined_prefixes: dict[str, str],
            known_prefixes: dict[str, str],
            query: str
    ) -> str:
        """
        Corrects the prefix definitions in the SPARQL query,
        which values differ from the declaration info available in the repository.
        :param defined_prefixes: prefixes defined in the SPARQL query
        :type defined_prefixes: dict[str, str]
        :param known_prefixes: prefixes from the declaration info available in the repository
        :type known_prefixes:  dict[str, str]
        :param query: SPARQL query
        :type query: str
        :return: updated SPARQL query, where the prefix definitions,
        which values differ from the declaration info available in the repository, are automatically corrected.
        The defined_prefixes are also updated.
        :rtype: str
        """
        for prefix, namespace in defined_prefixes.items():
            if prefix in known_prefixes and known_prefixes[prefix] != namespace:
                logging.debug(
                    f"Correcting wrong value of prefix {prefix} : {namespace} to {known_prefixes[prefix]}"
                )
                regex = re.compile("prefix\\s+%s:\\s*<%s>" % (prefix, namespace), flags=re.IGNORECASE)
                query = re.sub(regex, f"PREFIX {prefix}: <{known_prefixes[prefix]}>", query)
                defined_prefixes[prefix] = known_prefixes[prefix]
        return query

    @staticmethod
    def __get_prefixed_iris(query_part: str) -> set[tuple[str, str]]:
        """
        Returns the prefixed IRIs in the SPARQL query
        :param query_part: the query part of the parsed query
        :type query_part: str
        :return: prefixed IRIs in the SPARQL query as set of tuples (prefix, local name)
        :rtype: set[tuple[str, str]]
        """
        return set(map(
            lambda x: (x[0], x[1]),
            re.findall(r"pname_\{'prefix': '(.+?)', 'localname': '(.+?)'}", query_part)
        ))

    @staticmethod
    def __add_missing_prefixes(
            defined_prefixes: dict[str, str],
            known_prefixes: dict[str, str],
            prefixed_iris: set[tuple[str, str]],
            query: str
    ) -> str:
        """
        Adds prefixes used in the SPARQL query, which are not defined in it, but appear in
        the declaration info available in the repository
        :param defined_prefixes: prefixes defined in the SPARQL query
        :type defined_prefixes: dict[str, str]
        :param known_prefixes: prefixes from the declaration info available in the repository
        :type known_prefixes:  dict[str, str]
        :param prefixed_iris: the prefixed IRIs in the SPARQL query
        :type prefixed_iris: set[tuple[str, str]]
        :param query: SPARQL query
        :type query: str
        :return: updated SPARQL query, where missing prefixes are added automatically,
         if they appear in the declaration info available in the repository.
         The defined_prefixes are also updated.
        :rtype: str
        :raises ValueError, if the query uses prefixes, which are not defined in the query and don't appear in
        the declaration info available in the repository
        """
        references_prefixes = set(map(lambda x: x[0], prefixed_iris))
        if not references_prefixes.issubset(defined_prefixes.keys()):
            undefined_prefixes = references_prefixes - defined_prefixes.keys()
            for prefix in undefined_prefixes:
                if prefix in known_prefixes:
                    defined_prefixes[prefix] = known_prefixes[prefix]
                    query = f"PREFIX {prefix}: <{known_prefixes[prefix]}> " + query
            undefined_prefixes = undefined_prefixes - defined_prefixes.keys()
            if undefined_prefixes:
                raise ValueError(f"The following prefixes are undefined: {', '.join(undefined_prefixes)}")
        return query

    def __validate_iris_are_stored(
            self,
            defined_prefixes: dict[str, str],
            prefixed_iris: set[tuple[str, str]],
            query_part: str
    ) -> None:
        """
        Executes a SPARQL query, which uses the special predicate <http://www.ontotext.com/owlim/entity#id>
        to check if the IRIs in the SPARQL query are stored in the repository.
        :param defined_prefixes: prefixes defined in the SPARQL query
        :type defined_prefixes: dict[str, str]
        :param prefixed_iris: the prefixed IRIs in the SPARQL query
        :type prefixed_iris: set[tuple[str, str]]
        :param query_part: the query part of the parsed query
        :type query_part: str
        :rtype: None
        :raises ValueError, if one or more IRIs used in the query are not stored in repository
        """
        iris = self.__get_all_iris(defined_prefixes, prefixed_iris, query_part)
        iri_values = " ".join(map(lambda x: f"<{x}>", iris))
        iri_validation_query = """
        SELECT ?iri {
            VALUES ?iri {
                %s
            }
            ?iri <http://www.ontotext.com/owlim/entity#id> ?id .
            FILTER(?id < 0)
        }
        """
        result, _ = self.eval_sparql_query(iri_validation_query % iri_values, validation=False)
        invalid_iris = list(
            map(lambda x: f"<{x['iri']['value']}>", result["results"]["bindings"])
        )
        if invalid_iris:
            raise ValueError(
                f"The following IRIs are not used in the data stored in GraphDB: {', '.join(invalid_iris)}"
            )

    @staticmethod
    def __get_all_iris(
            defined_prefixes: dict[str, str],
            prefixed_iris: set[tuple[str, str]],
            query_part: str
    ) -> set[str]:
        """
        Returns all IRIs used in the SPARQL query
        :param defined_prefixes: prefixes defined in the SPARQL query
        :type defined_prefixes: dict[str, str]
        :param prefixed_iris: the prefixed IRIs in the SPARQL query
        :type prefixed_iris: set[tuple[str, str]]
        :param query_part: the query part of the parsed query
        :type query_part: str
        :return: all IRIs from the SPARQL query
        :rtype: set[str]
        """
        prefixed_iris_to_full_iris = {
            defined_prefixes[x[0]] + x[1]
            for x in prefixed_iris
        }
        full_iris = set(re.findall(r"rdflib\.term\.URIRef\('(.+?)'\)", query_part))
        iris = full_iris | prefixed_iris_to_full_iris
        iris = set(
            filter(
                lambda x: (
                              not x.startswith("http://www.w3.org/2001/XMLSchema#")
                          ) and (
                              not x.startswith("http://www.ontotext.com/owlim/RDFRank#")
                          ) and (
                              not x.startswith("http://www.ontotext.com/plugins/autocomplete#")
                          ) and (
                              not x.startswith("http://www.openrdf.org/schema/sesame#")
                          ) and (
                              not x.startswith("http://spinrdf.org/spif#")
                          ) and (
                              not x.startswith("http://www.ontotext.com/fts")
                          ),
                iris
            )
        )
        return iris

    def eval_sparql_query(
            self,
            query: str,
            result_format: str = None,
            validation: bool = True
    ) -> Tuple[Any, str]:
        """
        Executes the provided SPARQL query against GraphDB.

        :param query: the SPARQL query, which should be evaluated
        :type query: str
        :param result_format: Format of the results.
        Possible values are "json", "xml", "turtle", "n3", "rdf", "rdf+xml", "csv", "tsv", "json-ld"
        (defined as constants in SPARQLWrapper). All other cases are ignored.
        :type result_format: str
        :param validation: should be True, if the SPARQL query is generated from a LLM, and should be validated.
        The validation includes parsing of the query, checks for missing prefixes,
         or usage of IRIs, which are not stored in GraphDB.
        :type validation: bool
        :return: the results in the expected result_format
        :rtype:
        """
        if validation:
            query = self.__validate_query(query)

        self.__sparql_wrapper.setQuery(query)
        self.__sparql_wrapper.setMethod(Wrapper.POST)

        if result_format is None:
            if self.__sparql_wrapper.queryType in {"CONSTRUCT", "DESCRIBE"}:
                result_format = TURTLE
            else:
                result_format = JSON

        self.__sparql_wrapper.setReturnFormat(result_format)
        results = self.__sparql_wrapper.query().convert()
        if result_format != JSON:
            return results.decode("utf-8"), query
        else:
            return results, query



from abc import ABCMeta
from typing import Optional

from langchain_core.tools import BaseTool
from pydantic import model_validator
from typing_extensions import Self

from ttyg.graphdb import GraphDB


class BaseGraphDBTool(BaseTool, metaclass=ABCMeta):
    """Base tool for interacting with GraphDB"""

    graph: GraphDB
    """The GraphDB Client"""

    @property
    def min_graphdb_version(self) -> Optional[str]:
        """
        :return: the minimum GraphDB version required to use the tool
        :rtype: str
        """
        return None

    @model_validator(mode="after")
    def graphdb_version_compatibility(self) -> Self:
        if self.min_graphdb_version:
            graphdb_version = self.graph.version
            major, minor, _ = graphdb_version.split(".")
            major, minor = int(major), int(minor)

            min_major, min_minor = self.min_graphdb_version.split(".")
            min_major, min_minor = int(min_major), int(min_minor)

            if (major < min_major) or (major == min_major and minor < min_minor):
                raise ValueError(
                    f"GraphDB version {self.min_graphdb_version} or later required. Please, upgrade to a newer version."
                )

        return self



import logging
from functools import cached_property
from pathlib import Path
from typing import Optional

from langchain_core.callbacks import CallbackManagerForToolRun
from pydantic import model_validator, computed_field
from pyparsing import ParseException
from rdflib import Graph
from rdflib.plugins.sparql import prepareQuery
from typing_extensions import Self

from ttyg.utils import timeit
from .base import BaseGraphDBTool


class OntologySchemaAndVocabularyTool(BaseGraphDBTool):
    """
    Tool, which returns the configured ontology schema and vocabulary for the SPARQL queries.
    The ontology schema can be provided either with a path to a file in turtle format containing the ontology schema statements,
    or with a SPARQL CONSTRUCT query, which returns all the ontology schema statements.
    """

    name: str = "ontology_schema_and_vocabulary_tool"
    description: str = "Tool, which returns the configured ontology schema and vocabulary for the SPARQL queries"
    ontology_schema_file_path: Optional[Path] = None
    ontology_schema_query: Optional[str] = None

    @model_validator(mode="after")
    def valid_ontology_query(self) -> Self:
        """
        Validate the query is a valid SPARQL CONSTRUCT query
        """

        if self.ontology_schema_query:
            try:
                parsed_query = prepareQuery(self.ontology_schema_query)
            except ParseException as e:
                raise ValueError("Ontology query is not a valid SPARQL query.", e)

            if parsed_query.algebra.name != "ConstructQuery":
                raise ValueError(
                    "Invalid query type. Only CONSTRUCT queries are supported."
                )

        return self

    @model_validator(mode="after")
    def validate(self) -> Self:
        if self.ontology_schema_file_path and self.ontology_schema_query:
            raise ValueError(
                "Expected only one of ontology schema file path or ontology schema query. Both are provided."
            )
        if (not self.ontology_schema_file_path) and (not self.ontology_schema_query):
            raise ValueError(
                "Neither ontology schema file path nor ontology schema query is provided."
            )
        return self

    @computed_field
    @cached_property
    def schema_graph(self) -> Graph:
        if self.ontology_schema_query:
            logging.debug("Configuring the ontology schema with query.")
            sparql_results, _ = self.graph.eval_sparql_query(self.ontology_schema_query, validation=False)
            schema_graph = Graph().parse(
                data=sparql_results,
                format="turtle",
            )
            logging.debug(f"Collected {len(schema_graph)} ontology schema statements.")
            return schema_graph
        else:
            logging.debug("Configuring the ontology schema with file.")
            schema_graph = Graph().parse(
                data=self.ontology_schema_file_path.read_text(),
                format="turtle",
            )
            logging.debug(f"Collected {len(schema_graph)} ontology schema statements.")
            return schema_graph

    @timeit
    def _run(
            self,
            run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        return self.schema_graph.serialize(format="turtle")



import json
import logging
from typing import (
    Optional,
    Type,
    Tuple,
)

from langchain_core.callbacks import CallbackManagerForToolRun
from pydantic import BaseModel, Field

from ttyg.utils import timeit
from .base import BaseGraphDBTool


class SparqlQueryTool(BaseGraphDBTool):
    """
    Tool, which executes SPARQL queries generated by the agent.
    """

    class SearchInput(BaseModel):
        query: str = Field(
            description="A valid SPARQL SELECT, CONSTRUCT, DESCRIBE or ASK query without prefixes"
        )

    name: str = "sparql_query"
    description: str = "Query GraphDB by SPARQL SELECT, CONSTRUCT, DESCRIBE or ASK query and return result."
    response_format: str = "content_and_artifact"
    args_schema: Type[BaseModel] = SearchInput

    @timeit
    def _run(
            self,
            query: str,
            run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Tuple[str, str]:
        logging.debug(f"Executing generated SPARQL query {query}")
        query_results, actual_query = self.graph.eval_sparql_query(query)
        return json.dumps(query_results, indent=2), actual_query



from .graphdb_tools import (
    AutocompleteSearchTool,
    FTSTool,
    IRIDiscoveryTool,
    OntologySchemaAndVocabularyTool,
    RetrievalQueryTool,
    SimilaritySearchQueryTool,
    SparqlQueryTool
)
from .now_tool import NowTool

__all__ = [
    "AutocompleteSearchTool",
    "FTSTool",
    "IRIDiscoveryTool",
    "OntologySchemaAndVocabularyTool",
    "RetrievalQueryTool",
    "SimilaritySearchQueryTool",
    "SparqlQueryTool",
    "NowTool",
]



from .autocomplete_search_tool import AutocompleteSearchTool
from .fts_tool import FTSTool
from .iri_discovery_tool import IRIDiscoveryTool
from .ontology_schema_and_vocabulary_tool import OntologySchemaAndVocabularyTool
from .retrieval_query_tool import RetrievalQueryTool
from .similarity_search_query_tool import SimilaritySearchQueryTool
from .sparql_query_tool import SparqlQueryTool

__all__ = [
    "AutocompleteSearchTool",
    "FTSTool",
    "IRIDiscoveryTool",
    "OntologySchemaAndVocabularyTool",
    "RetrievalQueryTool",
    "SimilaritySearchQueryTool",
    "SparqlQueryTool",
]



import json
import logging
from typing import (
    Any,
    Optional,
    ClassVar,
    Type,
    Tuple,
)

from langchain_core.callbacks import CallbackManagerForToolRun
from pydantic import Field, model_validator, BaseModel
from typing_extensions import Self

from ttyg.graphdb import GraphDB, GraphDBRdfRankStatus
from ttyg.utils import timeit
from .base import BaseGraphDBTool


def _get_default_sparql_template(validated_data: dict[str, Any]) -> str:
    graph: GraphDB = validated_data["graph"]
    graphdb_version = graph.version
    major, minor, _ = graphdb_version.split(".")
    major, minor = int(major), int(minor)

    if major >= 10 and minor >= 8:
        return """PREFIX onto: <http://www.ontotext.com/>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX schema: <http://schema.org/>
PREFIX rank: <http://www.ontotext.com/owlim/RDFRank#>
SELECT ?iri ?label {{
    ?label onto:fts ("{query}" "*").
    ?iri rdfs:label|skos:prefLabel|schema:name ?label.
    ?iri rank:hasRDFRank ?rank .
}}
ORDER BY DESC(?rank)
LIMIT {limit}"""
    else:
        return """PREFIX onto: <http://www.ontotext.com/>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX schema: <http://schema.org/>
PREFIX rank: <http://www.ontotext.com/owlim/RDFRank#>
SELECT ?iri ?label {{
    ?label onto:fts ("{query}").
    ?iri rdfs:label|skos:prefLabel|schema:name ?label.
    ?iri rank:hasRDFRank ?rank .
}}
ORDER BY DESC(?rank)
LIMIT {limit}"""


class IRIDiscoveryTool(BaseGraphDBTool):
    """
    Tool, which uses GraphDB full-text search (FTS), to discover IRIs of concepts.
    The full-text search (FTS) must be enabled for the repository in order to use this tool.
    For details how to enable it check the documentation https://graphdb.ontotext.com/documentation/10.8/full-text-search.html#simple-full-text-search-index.
    It's also recommended to compute the RDF rank for the repository.
    For details how to compute it refer to the documentation https://graphdb.ontotext.com/documentation/10.8/ranking-results.html.
    The agent generates the fts search query, which is expanded in the SPARQL template.
    """

    class SearchInput(BaseModel):
        query: str = Field(description="FTS search query")

    min_graphdb_version: ClassVar[str] = "10.1"
    name: str = "iri_discovery"
    description: str = "Discovery IRIs by full-text search in labels."
    args_schema: Type[BaseModel] = SearchInput
    response_format: str = "content_and_artifact"
    query_template: str = Field(default_factory=lambda validated_data: _get_default_sparql_template(validated_data))
    limit: int = Field(default=10, ge=1)

    @model_validator(mode="after")
    def graphdb_config(self) -> Self:
        if not self.graph.fts_is_enabled():
            logging.warning(
                "You must enable the full-text search (FTS) index for the repository "
                "to use the IRI discovery tool."
            )

        rdf_rank_status = self.graph.get_rdf_rank_status()
        if rdf_rank_status != GraphDBRdfRankStatus.COMPUTED:
            logging.warning(
                f"The RDF Rank status of the repository is \"{rdf_rank_status.name}\". "
                f"It's recommended the status to be COMPUTED in order to use the IRI discovery tool."
            )

        return self

    @timeit
    def _run(
            self,
            query: str,
            run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Tuple[str, str]:
        query = self.query_template.format(query=query, limit=self.limit)
        logging.debug(f"Searching with iri discovery {query}")
        query_results, query = self.graph.eval_sparql_query(query, validation=False)
        return json.dumps(query_results, indent=2), query
